import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import tdsbrondata
from delta.tables import DeltaTable
from pyspark.sql import functions as F
from pyspark.sql.types import *
from functools import reduce

def initDeltaTable(tableName, columns, recreate):
    
    table = f"{tdsbrondata.schemaName}.{tableName}"
    tablePath = f"{tdsbrondata.tablesRootPath}/{tableName}"

    if recreate and DeltaTable.isDeltaTable(tdsbrondata._spark, tablePath):
        tdsbrondata._spark.sql(f"DROP TABLE IF EXISTS {table}")

    if not tdsbrondata._spark.catalog.tableExists(table):
        builder = DeltaTable.create(tdsbrondata._spark).tableName(tableName)

        for column in columns:
            builder = builder.addColumn(column["nameDelta"], dataType=column["type"])

        for column in tdsbrondata.scdColumns:
            builder = builder.addColumn(column["name"], dataType=column["type"])

        builder.execute()

    dt = DeltaTable.forPath(tdsbrondata._spark, tablePath)
    dt.vacuum(840)

    return dt

def applyFilters(df, filters):

    if not filters:
        return df

    conditions = reduce(
        lambda accumulator, f: accumulator & (F.col(f[0]) == f[2]) if f[1] == "==" else accumulator & (F.col(f[0]) != f[2]),
        filters,
        F.lit(True)
    )
    
    return df.filter(conditions)

def reduceExistingData(df, columns):
    
    dfCurrent = df.filter((F.col("CurrentFlag") == 1) & (F.col("ScdEndDate").isNull()))

    columnsSelected = [column["nameDelta"] for column in columns]

    return dfCurrent.select(*columnsSelected)

def reduceNewData(df, columns):
    
    columnsSelected = [column["nameSpark"] for column in columns]
    
    renameMapping = {column["nameSpark"]: column["nameDelta"] for column in columns}
    
    dfReduced = df.select(*columnsSelected)
    
    for nameOld, nameNew in renameMapping.items():
        dfReduced = dfReduced.withColumnRenamed(nameOld, nameNew)

    return dfReduced.distinct()

def getSourceData(tableName, sourceDataOptions):
    if tdsbrondata.sourceDataMode == 'automatic':
        return getAutomaticSourceData(tableName=tableName)
    elif tdsbrondata.sourceDataMode == 'manual':
        return getManualSourceData(tableName=tableName, sourceDataOptions=sourceDataOptions)

def getAutomaticSourceData(tableName):

    filePathFull = getAutomaticSourceDataPath(tableName, 'full')
    filePathIncremental = getAutomaticSourceDataPath(tableName, 'incremental')
    
    if tdsbrondata._notebookutils.fs.exists(filePathFull):
        return tdsbrondata._spark.read.parquet(filePathFull)
    elif tdsbrondata._notebookutils.fs.exists(filePathIncremental):
        return tdsbrondata._spark.read.parquet(filePathIncremental)
    else:
        raise FileNotFoundError(f"No automatic source data file found for table '{tableName}'.")

def getManualSourceData(tableName, sourceDataOptions):

    filePath = getManualSourceDataPath(tableName)
    
    if tdsbrondata._notebookutils.fs.exists(filePath):

        reader = tdsbrondata._spark.read
        
        if sourceDataOptions:
            for option, value in sourceDataOptions:
                reader = reader.option(option, value)
        
        return reader.csv(filePath)
    else:
        raise FileNotFoundError(f"No manual source data file found for table '{tableName}'.")
    
def getAutomaticSourceDataPath(tableName, loadType):
    path = f"{tdsbrondata.automaticDataPath}/{tableName}_{loadType}.parquet"
    path = path.replace(tdsbrondata.lakehouseNameSilver, tdsbrondata.lakehouseNameBronze)
    path = path.replace(tdsbrondata.schemaNameSilver, tdsbrondata.schemaNameBronze)
    return path

def getManualSourceDataPath(tableName):
    path = f"{tdsbrondata.manualDataPath}/{tdsbrondata.sourceDataPeriod}_{tableName}.csv"
    path = path.replace(tdsbrondata.lakehouseNameSilver, tdsbrondata.lakehouseNameBronze)
    path = path.replace(tdsbrondata.schemaNameSilver, tdsbrondata.schemaNameBronze)
    return path

def extractMutations(sourceData, columns, dtExisting, filters):

    dfExisting = dtExisting.toDF()
    dfExistingReduced = reduceExistingData(df=dfExisting, columns=columns)

    dfNew = applyFilters(df=sourceData, filters=filters)
    dfNewReduced = reduceNewData(df=dfNew, columns=columns)

    for columnName, dtype in dfExistingReduced.dtypes:
        dfNewReduced = dfNewReduced.withColumn(columnName, F.col(columnName).cast(dtype))

    dfMutations = dfNewReduced.exceptAll(dfExistingReduced)

    return dfMutations

def extractDeletions(sourceData, columns, primaryKey, dtExisting, filters):

    isCompositeKey = isinstance(primaryKey, list)

    dfExisting = dtExisting.toDF()
    dfExistingReduced = reduceExistingData(df=dfExisting, columns=columns)

    dfNew = applyFilters(df=sourceData, filters=filters)
    dfNewReduced = reduceNewData(df=dfNew, columns=columns)

    for columnName, dtype in dfExistingReduced.dtypes:
        dfNewReduced = dfNewReduced.withColumn(columnName, F.col(columnName).cast(dtype))

    if isCompositeKey:
        joinCondition = [dfExistingReduced[key] == dfNewReduced[key] for key in primaryKey]
    else:
        joinCondition = dfExistingReduced[primaryKey] == dfNewReduced[primaryKey]

    dfDeletions = dfExistingReduced.join(dfNewReduced, on=joinCondition, how="left_anti")

    return dfDeletions

def addSurrogateKeys(dtExisting, dfMutations):
    
    maxSurrogateKey = dtExisting.toDF().agg({"SurrogateKey": "max"}).collect()[0][0] or 0

    dfMutations = dfMutations.withColumn(
        "SurrogateKey",
        F.monotonically_increasing_id() + maxSurrogateKey + 1
    )

    return dfMutations

def updateRecords(columns, primaryKey, dtExisting, dfMutations):

    isCompositeKey = isinstance(primaryKey, list)

    if isCompositeKey:
        mergeCondition = " AND ".join([f"existing.{key} = mutations.{key}" for key in primaryKey]) + " AND existing.CurrentFlag = true"
    else:
        mergeCondition = f"existing.{primaryKey} = mutations.{primaryKey} AND existing.CurrentFlag = true"
    
    updateCondition = " OR ".join([
        f"NOT (existing.{column['nameDelta']} <=> mutations.{column['nameDelta']})"
        for column in columns
    ])

    (
        dtExisting.alias("existing")
        .merge(
            dfMutations.alias("mutations"),
            mergeCondition
        )
        .whenMatchedUpdate(
            condition=F.expr(updateCondition),
            set={
                "ScdEndDate": F.current_timestamp(),
                "CurrentFlag": F.lit(False)
            }
        )
        .execute()
    )

def insertRecords(columns, primaryKey, dtExisting, dfMutations):

    isCompositeKey = isinstance(primaryKey, list)

    if isCompositeKey:
        mergeCondition = " AND ".join([f"existing.{key} = mutations.{key}" for key in primaryKey]) + " AND existing.CurrentFlag = true"
    else:
        mergeCondition = f"existing.{primaryKey} = mutations.{primaryKey} AND existing.CurrentFlag = true"

    insertValues = {
        column["nameDelta"]: (
            F.col(f"mutations.{column['nameDelta']}")
        )
        for column in columns
    }

    insertValues.update({
        "SurrogateKey": F.col("mutations.SurrogateKey"),
        "CurrentFlag": F.lit(True),
        "ScdStartDate": F.current_timestamp(),
        "ScdEndDate": F.lit(None).cast("timestamp")
    })

    dtExisting.alias("existing") \
        .merge(
            dfMutations.alias("mutations"),
            mergeCondition
        ) \
        .whenNotMatchedInsert(
            values=insertValues
        ) \
        .execute()

def deleteRecords(primaryKey, dtExisting, dfDeletions):

    isCompositeKey = isinstance(primaryKey, list)

    if isCompositeKey:
        mergeCondition = " AND ".join([f"existing.{key} = deletions.{key}" for key in primaryKey]) + " AND existing.CurrentFlag = true"
    else:
        mergeCondition = f"existing.{primaryKey} = deletions.{primaryKey} AND existing.CurrentFlag = true"

    dtExisting.alias("existing") \
        .merge(
            dfDeletions.alias("deletions"),
            mergeCondition
        ) \
        .whenMatchedUpdate(
            set={
                "ScdEndDate": F.current_timestamp(),
                "CurrentFlag": F.lit(False)
            }
        ) \
        .execute()

def mutateDeltaTable(columns, primaryKey, dtExisting, dfMutations, dfDeletions):

    updateRecords(columns=columns, primaryKey=primaryKey, dtExisting=dtExisting, dfMutations=dfMutations)
    insertRecords(columns=columns, primaryKey=primaryKey, dtExisting=dtExisting, dfMutations=dfMutations)
    deleteRecords(primaryKey=primaryKey, dtExisting=dtExisting, dfDeletions=dfDeletions)

def processData(tableName, columns, primaryKey, filters, recreate, sourceDataOptions=[], sourceDataOverride=None):

    if sourceDataOverride is not None:
        sourceData = sourceDataOverride
    else:
         sourceData = getSourceData(tableName=tableName, sourceDataOptions=sourceDataOptions)

    sourceData = sourceData.toDF(*[column.replace(" ", "_") for column in sourceData.columns])
    
    dtExisting = initDeltaTable(tableName=tableName, columns=columns, recreate=recreate)

    dfMutations = extractMutations(sourceData=sourceData, columns=columns, dtExisting=dtExisting, filters=filters)
    dfMutations = addSurrogateKeys(dtExisting=dtExisting, dfMutations=dfMutations)

    dfDeletions = extractDeletions(sourceData=sourceData, columns=columns, primaryKey=primaryKey, dtExisting=dtExisting, filters=filters)

    mutateDeltaTable(columns=columns, primaryKey=primaryKey, dtExisting=dtExisting, dfMutations=dfMutations, dfDeletions=dfDeletions)


import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import tdsbrondata

def modules(notebookutils, spark):
    tdsbrondata._notebookutils = notebookutils
    tdsbrondata._spark = spark
    tdsbrondata._spark.conf.set("spark.sql.execution.arrow.pyspark.enabled", "false")
    tdsbrondata._spark.conf.set("spark.sql.parquet.vorder.enabled", "true")
    tdsbrondata._spark.conf.set("spark.microsoft.delta.optimizeWrite.enabled", "true")
    tdsbrondata._spark.conf.set("spark.microsoft.delta.optimizeWrite.binSize", "134217728")
    tdsbrondata._spark.conf.set("spark.databricks.delta.retentionDurationCheck.enabled", "false")

def lakehouse(schemaName, sourceDataMode, sourceDataPeriod):
    tdsbrondata.schemaName = schemaName
    tdsbrondata.schemaNameBronze = ('B' + schemaName[1:] if schemaName.startswith('S') else schemaName)
    tdsbrondata.schemaNameSilver = ('S' + schemaName[1:] if schemaName.startswith('B') else schemaName)
    tdsbrondata.sourceDataMode = sourceDataMode
    tdsbrondata.sourceDataPeriod = sourceDataPeriod
    tdsbrondata.workspaceName = workspaceName = tdsbrondata._spark.conf.get("trident.workspace.name", "")
    tdsbrondata.lakehouseName = lakehouseName = tdsbrondata._spark.conf.get("trident.lakehouse.name", "")
    tdsbrondata.lakehouseNameBronze = ('B' + lakehouseName[1:] if lakehouseName.startswith('S') else lakehouseName)
    tdsbrondata.lakehouseNameSilver = ('S' + lakehouseName[1:] if lakehouseName.startswith('B') else lakehouseName)
    tdsbrondata.automaticDataPath = f"abfss://{workspaceName}@onelake.dfs.fabric.microsoft.com/{lakehouseName}.Lakehouse/Files/AutomaticData/{schemaName}/{sourceDataPeriod}"
    tdsbrondata.manualDataPath = f"abfss://{workspaceName}@onelake.dfs.fabric.microsoft.com/{lakehouseName}.Lakehouse/Files/ManualData/{schemaName}"
    tdsbrondata.tablesRootPath = f"abfss://{workspaceName}@onelake.dfs.fabric.microsoft.com/{lakehouseName}.Lakehouse/Tables/{schemaName}"

    fullNamespace = f"{workspaceName}.{lakehouseName}.{schemaName}"
    if fullNamespace in [row['namespace'] for row in tdsbrondata._spark.sql("SHOW SCHEMAS").collect()]:
        tdsbrondata._spark.sql(f"USE DATABASE {lakehouseName}.{schemaName}")

def keyvault(keyvaultName):
    tdsbrondata.keyvaultName = keyvaultName
    tdsbrondata.keyvaultUrl = f"https://{keyvaultName}.vault.azure.net/"


import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import tdsbrondata
from delta.tables import DeltaTable
from pyspark.sql import functions as F
from pyspark.sql.types import *
from datetime import datetime

def writeToParquet(df, name, loadType):
    filePath = f"{tdsbrondata.automaticDataPath}/{name}_{loadType}.parquet"

    for field in df.schema.fields:
        if field.dataType.simpleString() == "void":
            df = df.withColumn(field.name, F.col(field.name).cast(StringType()))

    for loadType in ["full", "incremental"]:
        existingFilePath = f"{tdsbrondata.automaticDataPath}/{name}_{loadType}.parquet"
        if tdsbrondata._notebookutils.fs.exists(existingFilePath):
            tdsbrondata._notebookutils.fs.rm(existingFilePath, recurse=True)

    df.write.parquet(filePath)

def writeToTable(df, name, mode):

    overwriteSchema = (mode == 'overwrite')

    tablePath = f"{tdsbrondata.tablesRootPath}/{name}"
    df.write.format("delta").mode(mode).option("overwriteSchema", overwriteSchema).save(tablePath)

def upsertTable(df, name, compositeKeyColumns):

    condition = " AND ".join([f"target.{column} = source.{column}" for column in compositeKeyColumns])

    tablePath = f"{tdsbrondata.tablesRootPath}/{name}"
    deltaTable = DeltaTable.forPath(tdsbrondata._spark, tablePath)

    deltaTable.alias("target").merge(df.alias("source"), condition).whenMatchedUpdateAll().whenNotMatchedInsertAll().execute()

def getMostRecentParquetDate(name):
    sourceDataDirectory = '/'.join(tdsbrondata.automaticDataPath.rsplit('/', 1)[:-1])
    directories = tdsbrondata._notebookutils.fs.ls(sourceDataDirectory)

    dateDirectories = []
    for directory in directories:
        try:
            directoryDate = datetime.strptime(directory.name, '%Y%m%d')
            dateDirectories.append((directoryDate, directory.name))
        except ValueError:
            continue

    dateDirectories.sort(reverse=True, key=lambda x: x[0])

    if not dateDirectories:
        return None

    today = datetime.now().date()

    for date, directoryName in dateDirectories:
        if date.date() < today:
            mostRecentDirectory = f"{sourceDataDirectory}/{directoryName}"
            files = tdsbrondata._notebookutils.fs.ls(mostRecentDirectory)

            for file in files:
                if (file.name == f"{name}_full.parquet" or file.name == f"{name}_incremental.parquet"):
                    return date.strftime('%Y-%m-%d 00:00:00')

    return None



from pyspark.sql.types import *

_notebookutils = None
_spark = None

schemaName = None
schemaNameBronze = None
schemaNameSilver = None
workspaceName = None
lakehouseName = None
lakehouseNameBronze = None
lakehouseNameSilver = None
keyvaultName = None
keyvaultUrl = None

sourceDataMode = None
sourceDataPeriod = None

automaticDataPath = None
manualDataPath = None
tablesRootPath = None

typeMapping = {
    "BINARY": BinaryType(),
    "BOOLEAN": BooleanType(),
    "BYTE": ByteType(),
    "DATE": DateType(),
    "DOUBLE": DoubleType(),
    "FLOAT": FloatType(),
    "INTEGER": IntegerType(),
    "LONG": LongType(),
    "STRING": StringType(),
    "TIMESTAMP": TimestampType()
}

scdColumns = [
    {"name": "SurrogateKey", "type": "LONG"},
    {"name": "CurrentFlag", "type": "BYTE"},
    {"name": "ScdStartDate", "type": "TIMESTAMP"},
    {"name": "ScdEndDate", "type": "TIMESTAMP"}
]


import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import tdsbrondata
from pyspark.sql import functions as F
from pyspark.sql.types import *

def afasDebtor(dataToCheck, lakehouseName, schemaName, tableName, afasDebtors, idColumn):

    dataToCheck = (
        dataToCheck
        .filter((F.col("CurrentFlag") == 1) & (F.col("ScdEndDate").isNull()))
        .withColumn(
            "verkooprelatie_id",
            F.trim(F.lower(F.col("verkooprelatie_id").cast("string")))
        )
        .alias("check")
    )

    afasDebtors = (
        afasDebtors
        .select("Verkooprelatie_id", "Geblokkeerd_voor_levering", "Volledig_blokkeren")
        .withColumn(
            "verkooprelatie_id",
            F.trim(F.lower(F.col("Verkooprelatie_id").cast("string")))
        )
        .alias("afas")
    )

    dienst_id = schemaName[1:].split('_')[0]

    result = (
        dataToCheck.join(afasDebtors, F.col("check.verkooprelatie_id") == F.col("afas.verkooprelatie_id"), how="left")
        .withColumn(
            "remarks",
            F.when(F.col("check.verkooprelatie_id").isNull(), "AFAS verkooprelatie leeg")
            .when(F.col("afas.verkooprelatie_id").isNull(), "AFAS verkooprelatie niet gevonden")
            .when((F.col("afas.Geblokkeerd_voor_levering") == True) | (F.col("afas.Volledig_blokkeren") == True), "AFAS verkooprelatie geblokkeerd")
        )
        .filter(F.col("remarks").isNotNull())
        .withColumn("lakehouseName", F.lit(lakehouseName))
        .withColumn("schemaName", F.lit(schemaName))
        .withColumn("tableName", F.lit(tableName))
        .withColumn("idColumn", F.lit(idColumn))
        .withColumn("id", F.col(f"check.{idColumn}"))
        .withColumn("surrogateKey", F.col("check.SurrogateKey"))
        .withColumn("dienst_id", F.lit(dienst_id))
        .select("lakehouseName", "schemaName", "tableName", "idColumn", "id", "surrogateKey", "dienst_id", F.col("check.verkooprelatie_id"), "remarks")
    )

    return result


import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import tdsbrondata
from pyspark.sql import functions as F
from pyspark.sql.types import *
from delta.tables import DeltaTable

requiredColumns = {
    "DienstId": IntegerType(),
    "Source": StringType(),
    "SurrogateKey": LongType(),
    "FacturatieMaand": StringType(),
    "VerkooprelatieId": StringType(),
    "ItemCode": StringType(),
    "Aantal": FloatType()
}

optionalColumns = {
    "AfwijkendePrijs": FloatType(),
    "DatumVanOrigineel": StringType(),
    "DatumTotOrigineel": StringType(),
    "AantalOrigineel": FloatType(),
    "DurationOrigineel": FloatType()
}

def validateItemsSchema(items):
    schemaDict = {f.name: f.dataType for f in items.schema.fields}

    for columnName, expectedType in requiredColumns.items():
        if columnName not in schemaDict:
            raise ValueError(f"Missing required column '{columnName}'")
        if type(schemaDict[columnName]) != type(expectedType):
            raise TypeError(f"Column '{columnName}' has type {schemaDict[columnName]}, expected {expectedType}")

    for columnName, expectedType in optionalColumns.items():
        if columnName in schemaDict and type(schemaDict[columnName]) != type(expectedType):
            raise TypeError(f"Column '{columnName}' has type {schemaDict[columnName]}, expected {expectedType}")

def writeItems(items):
    validateItemsSchema(items)

    table = "items"
    tablePath = f"{tdsbrondata.tablesRootPath}/{table}"

    if DeltaTable.isDeltaTable(tdsbrondata._spark, tablePath):
        deltaTable = DeltaTable.forPath(tdsbrondata._spark, tablePath)
        facturatieMaand = items.select("FacturatieMaand").head()["FacturatieMaand"]
        deltaTable.delete(f"FacturatieMaand = '{facturatieMaand}'")
        items.write.format("delta").mode("append").save(tablePath)
    else:
        items.write.format("delta").mode("overwrite").option("overwriteSchema", "true").save(tablePath)

    writeLines(items)

def writeLines(items):

    table = "lines"
    tablePath = f"{tdsbrondata.tablesRootPath}/{table}"

    facturatieMaand = items.select("FacturatieMaand").head()["FacturatieMaand"]

    existingOptionalColumns = [col for col in optionalColumns.keys() if col in items.columns]

    if existingOptionalColumns:
        filterExpression = " AND ".join([f"{col} IS NULL" for col in existingOptionalColumns])
        itemsToAggregate = items.filter(filterExpression)
        itemsToPreserve = items.filter(f"NOT ({filterExpression})")
    else:
        itemsToAggregate = items
        itemsToPreserve = items.limit(0)

    for col, colType in optionalColumns.items():
        if col not in items.columns:
            itemsToAggregate = itemsToAggregate.withColumn(col, F.lit(None).cast(colType))
            itemsToPreserve = itemsToPreserve.withColumn(col, F.lit(None).cast(colType))

    aggregated = itemsToAggregate.groupBy(
        "DienstId", "FacturatieMaand", "VerkooprelatieId", "ItemCode"
    ).agg(
        F.sum("Aantal").alias("Aantal"),
        F.collect_list("SurrogateKey").alias("SurrogateKeys")
    )

    for col, colType in optionalColumns.items():
        aggregated = aggregated.withColumn(col, F.lit(None).cast(colType))

    aggregated = aggregated.withColumn(
        "HasItems",
        F.when(
            (F.size("SurrogateKeys") > 1)
            & (F.expr("aggregate(SurrogateKeys, true, (accumulator, x) -> accumulator AND x IS NOT NULL)")),
            True
        ).otherwise(False)
    ).drop("SurrogateKeys")

    itemsToPreserve = itemsToPreserve.withColumn(
        "HasItems",
        F.when(F.col("SurrogateKey").isNotNull(), True).otherwise(False)
    )

    for col in ["Source", "SurrogateKey"]:
        if col not in aggregated.columns:
            aggregated = aggregated.withColumn(col, F.lit(None).cast(items.schema[col].dataType))

    aggregated = aggregated.select(itemsToPreserve.columns)

    lines = aggregated.unionByName(itemsToPreserve)

    if DeltaTable.isDeltaTable(tdsbrondata._spark, tablePath):
        deltaTable = DeltaTable.forPath(tdsbrondata._spark, tablePath)
        deltaTable.delete(f"FacturatieMaand = '{facturatieMaand}'")
        lines.write.format("delta").mode("append").save(tablePath)
    else:
        lines.write.format("delta").mode("overwrite").option("overwriteSchema", "true").save(tablePath)



import sys
import os

sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import tdsbrondata
from pyspark.sql import functions as F
from pyspark.sql.types import *

def getSecret(secretName):
    return tdsbrondata._notebookutils.credentials.getSecret(tdsbrondata.keyvaultUrl, secretName)

def getCurrentData(workspaceName, lakehouseName, schemaName, tableName, usesScd):
    
    abfss = f"abfss://{workspaceName}@onelake.dfs.fabric.microsoft.com/{lakehouseName}.Lakehouse/Tables/{schemaName}/{tableName}"

    data = tdsbrondata._spark.read.format("delta").load(abfss)
    
    if(usesScd):
        data = data.filter(
            (F.col("CurrentFlag") == 1) & 
            (F.col("ScdEndDate").isNull())
        )

    return data

def addEventLogEntry(lakehouseName, schemaName, tableName, eventType, eventResult):

    abfss = 'abfss://Tosch-Data@onelake.dfs.fabric.microsoft.com/M_Metadata.Lakehouse/Tables/dbo'
    
    schema = StructType([
        StructField("LakehouseName", StringType(), True),
        StructField("SchemaName", StringType(), True),
        StructField("TableName", StringType(), True),
        StructField("EventType", StringType(), True),
        StructField("EventResult", StringType(), True)
    ])

    row = {"LakehouseName": lakehouseName, "SchemaName": schemaName, "TableName": tableName, "EventType": eventType, "EventResult": eventResult}
    df = tdsbrondata._spark.createDataFrame([row], schema=schema)

    try:
        dfExisting = tdsbrondata._spark.read.format("delta").load(f"{abfss}/eventlog")
    except Exception as e:
        dfExisting = tdsbrondata._spark.createDataFrame([], schema=schema)

    isDuplicate = (
        dfExisting.filter(
            (F.col("LakehouseName") == row["LakehouseName"]) &
            (F.col("SchemaName") == row["SchemaName"]) &
            (F.col("TableName") == row["TableName"]) &
            (F.col("EventType") == row["EventType"]) &
            (F.col("EventResult") == row["EventResult"])
        )
        .limit(1)
        .count() > 0
    )

    if not isDuplicate:
        df.write.mode("append").format("delta").save(f"{abfss}/eventlog")

def fileToBeProcessed(lakehouseName, schemaName, tableName, filePattern):

    shouldBeProcessed = True

    eventLog = tdsbrondata._spark.read.format("delta").load("abfss://Tosch-Data@onelake.dfs.fabric.microsoft.com/M_Metadata.Lakehouse/Tables/dbo/eventlog")

    fileAdded = eventLog.filter(
        (F.col("LakehouseName") == lakehouseName) & 
        (F.col("SchemaName") == schemaName) & 
        (F.col("TableName") == tableName) & 
        (F.col("EventType") == "fileAdded") & 
        (F.col("EventResult").startswith(filePattern))
    ).limit(1)

    fileAdded = fileAdded.collect()
    if fileAdded:
        fileName = fileAdded[0]["EventResult"]
        fileProcessed = eventLog.filter(
            (F.col("LakehouseName") == lakehouseName) & 
            (F.col("SchemaName") == schemaName) & 
            (F.col("TableName") == tableName) & 
            (F.col("EventType") == "fileProcessed") & 
            (F.col("EventResult").startswith(filePattern))
        ).limit(1)

        fileProcessed = fileProcessed.collect()
        if fileProcessed:
            shouldBeProcessed = False
    else:
        shouldBeProcessed = False

    if shouldBeProcessed:
        return fileName
    else:
        return None